from abc import ABC, abstractmethod
from typing import Any

from sapl_base.decision import Decision


class ConstraintHandlerProvider(ABC):
    """
    Baseclass of a ConstraintHandlerProvider
    """

    @abstractmethod
    def priority(self) -> int:
        """
        ConstraintHandlerProvider are sorted by the value of the priority, when the ConstraintHandlerBundle is created

        :return: value by which ConstraintHandlerProvider are sorted
        """
        return 0

    @abstractmethod
    def is_responsible(self, constraint) -> bool:
        """
        Determine if this ConstraintHandler is responsible for the provided constraint

        :param constraint: A constraint, which can be an Obligation or an Advice, for which the
        ConstraintHandlerProvider checks if it is responsible to handle it.
        :return: Is this ConstraintHandlerProvider responsible to handle the provided constraint
        """
        pass

    @abstractmethod
    def handle(self, argument):
        """
        Abstractmethod, which needs to be implemented by a ConstraintHandlerProvider
        :param argument: The argument, which is provided to the ConstraintHandler, when it is called. This argument can
        be an Exception, function, decision, or the result of the executed function.
        """
        pass


class ErrorConstraintHandlerProvider(ConstraintHandlerProvider, ABC):
    """
    A Class, which can be inherited to create a ConstraintHandlerProvider to handle Exceptions,
    which are thrown, when an error occurs while enforcing a decorated function or class
    """

    @abstractmethod
    def handle(self, exception: Exception) -> Exception:
        """
        Abstractmethod, which needs to be implemented by an ErrorConstraintHandlerProvider, which will be called by
        the ConstraintHandlerBundle if the ErrorConstraintHandlerProvider is responsible for a given constraint

        :param exception: Exception which should be handled
        :return: An Exception, which will be provided as Argument
        for the next ErrorConstraintHandlerProvider, or raised after when there is no next
        ErrorConstraintHandlerProvider
        """
        pass




class OnDecisionConstraintHandlerProvider(ConstraintHandlerProvider, ABC):
    """
    A Class, which can be inherited to create a ConstraintHandlerProvider which can handle a Decision,
    which is received from the PDP
    """

    @abstractmethod
    def handle(self, decision: Decision) -> None:
        """
        Abstractmethod, which needs to be implemented by an OnDecisionHandlerProvider, which will be called by
        the ConstraintHandlerBundle if the OnDecisionHandlerProvider is responsible for a given constraint

        :param decision: Decision which should be handled
        pass
        """





class FunctionArgumentsConstraintHandlerProvider(ConstraintHandlerProvider, ABC):
    """
    A Class, which can be inherited to create a ConstraintHandlerProvider to handle obligations and advices
    regarding the arguments, which with the enforced function will be called
    """

    @abstractmethod
    def handle(self, function_arguments: dict) -> dict:
        """
        Abstractmethod, which needs to be implemented by an FunctionArgumentsConstraintHandlerProvider, which will be
        called by the ConstraintHandlerBundle if the FunctionArgumentsConstraintHandlerProvider is responsible for a
        given constraint

        :param function_arguments: a dict containing the arguments, with which the enforced function will be called
        :return: a dict, with which the enforced function will be called
        """
        pass


class ResultConstraintHandlerProvider(ConstraintHandlerProvider, ABC):
    """
    A Class, which can be inherited to create a ConstraintHandlerProvider to handle obligations and advices regarding
    the result of the execution of the enforced function.
    """

    @abstractmethod
    def handle(self, result: Any) -> Any:
        """
        Abstractmethod, which needs to be implemented by an ResultConstraintHandlerProvider, which will be called by
        the ConstraintHandlerBundle if the ResultConstraintHandlerProvider is responsible for a given constraint

        :param result: the result of the enforced function, which was called
        :return: an object, which is treated as if it were the result of the execution of the enforced function
        """
        pass



from functools import wraps


class NoHandlerException(Exception):
    """Raised when an object, which is not a ConstraintHandler is added to the ConstraintHandlerService"""
    pass


def double_wrap(f):
    """
    a decorator, allowing the decorator to be used as:
    @decorator(with, arguments, and=kwargs) or @decorator

    :type f: function or method
    :param f: function or method use the decorator
    """

    @wraps(f)
    def new_dec(*args, **kwargs):
        if len(args) == 1 and len(kwargs) == 0 and callable(args[0]):
            # actual decorated fn
            return f(args[0])
        else:
            # decorator arguments
            return lambda real_fn: f(real_fn, *args, **kwargs)

    return new_dec





from sapl_base.constraint_handling.constraint_handler_bundle import ConstraintHandlerBundle
from sapl_base.constraint_handling.constraint_handler_provider import OnDecisionConstraintHandlerProvider, \
    ErrorConstraintHandlerProvider, FunctionArgumentsConstraintHandlerProvider, ResultConstraintHandlerProvider, \
    ConstraintHandlerProvider
from sapl_base.decision import Decision
from sapl_base.policy_enforcement_points.policy_enforcement_point import permission_denied_exception


class ConstraintHandlerService:
    """
    Class, which contains all available ConstraintHandlerProvider and creates ConstraintHandlerBundle with
    responsible ConstraintHandlerProvider to handle the Obligations and Advices of a Decision
    """


    def __init__(self):
        self.on_decision_handler : list[OnDecisionConstraintHandlerProvider]= []
        self.error_handler : list[ErrorConstraintHandlerProvider]= []
        self.result_handler : list[ResultConstraintHandlerProvider]= []
        self.function_arguments_mapper : list[FunctionArgumentsConstraintHandlerProvider]= []

    def register_decision_constraint_handler_provider(self, providers:
    list[OnDecisionConstraintHandlerProvider] | OnDecisionConstraintHandlerProvider):
        try:
            for provider in providers:
                self.on_decision_handler.append(provider)
        except TypeError:
            self.on_decision_handler.append(providers)

    def register_error_constraint_handler_provider(self, providers: list[ErrorConstraintHandlerProvider] | ErrorConstraintHandlerProvider):
        try:
            for provider in providers:
                self.error_handler.append(provider)
        except TypeError:
            self.error_handler.append(providers)

    def register_result_constraint_handler_provider(self, providers: list[ResultConstraintHandlerProvider] | ResultConstraintHandlerProvider):
        try:
            for provider in providers:
                self.result_handler.append(provider)
        except TypeError:
            self.result_handler.append(providers)

    def register_function_arguments_constraint_handler_provider(self, providers: list[
        FunctionArgumentsConstraintHandlerProvider] | FunctionArgumentsConstraintHandlerProvider):
        try:
            for provider in providers:
                self.function_arguments_mapper.append(provider)
        except TypeError:
            self.function_arguments_mapper.append(providers)

    def build_post_enforce_bundle(self, decision: Decision) -> ConstraintHandlerBundle:
        """
        Create a ConstraintHandlerBundle for a function, which was decorated with @post_enforce, or
        @pre_and_post_enforce to handle the obligations and advices of the decision

        :param decision: decision from the PolicyDecisionPoint
        :raise PermissionDenied Exception, when at least one obligation can't be handled:
        :return: A ConstraintHandlerBundle, containing all ConstraintHandlerProvider, which are responsible for the
        obligations and advices of the decision

        """

        unhandled_obligations = []
        for obligation in decision.obligations:
            unhandled_obligations.append(obligation)
        on_decision_handler, error_handler, result_handler = self._build_basic_bundle(decision.obligations,
                                                                                      decision.advice,
                                                                                      unhandled_obligations)
        if unhandled_obligations:
            raise permission_denied_exception
        return ConstraintHandlerBundle(on_decision_handler, error_handler, result_handler)

    def build_pre_enforce_bundle(self, decision: Decision) -> ConstraintHandlerBundle:
        """
        Create a ConstraintHandlerBundle for a function, which was decorated with @pre_enforce, or
        @pre_and_post_enforce to handle the obligations and advices of the decision

        :param decision: decision from the PolicyDecisionPoint
        :raise PermissionDenied Exception, when at least one obligation can't be handled:
        :return: A ConstraintHandlerBundle, containing all ConstraintHandlerProvider, which are responsible for the
        obligations and advices of the decision
        """

        unhandled_obligations = []
        for obligation in decision.obligations:
            unhandled_obligations.append(obligation)
        on_decision_handler, error_handler, result_handler = self._build_basic_bundle(decision.obligations,
                                                                                      decision.advice,
                                                                                      unhandled_obligations)

        function_arguments_mapper = self._create_function_argument_mapper(decision.obligations, unhandled_obligations)
        function_arguments_mapper.extend(self._create_function_argument_mapper(decision.advice, []))
        if unhandled_obligations:
            raise permission_denied_exception
        return ConstraintHandlerBundle(on_decision_handler, error_handler, result_handler, function_arguments_mapper)

    def _build_basic_bundle(self, obligations: list, advices: list, unhandled_obligations: list) \
            -> tuple[list[OnDecisionConstraintHandlerProvider], list[ErrorConstraintHandlerProvider],
            list[ResultConstraintHandlerProvider]]:
        """
        Create lists of ConstraintHandlerProvider which are responsible for the given obligations and advices

        :param obligations: a constraint, which must be handled by a ConstraintHandlerProvider
        :param advices: a constraint which should be handled by a ConstraintHandlerProvider
        :param unhandled_obligations: list of all obligations, for which no responsible ConstraintHandlerProvider has been found
        :return: Tuple of three lists with ConstraintHandlerProvider
        """
        on_decision_handler = self._create_on_decision_handler(obligations, unhandled_obligations)
        on_decision_handler.extend(self._create_on_decision_handler(advices, []))

        error_handler = self._create_on_error_handler(obligations, unhandled_obligations)
        error_handler.extend(self._create_on_error_handler(advices, []))

        result_handler = self._create_result_handler(obligations, unhandled_obligations)
        result_handler.extend(self._create_result_handler(advices, []))

        return on_decision_handler, error_handler, result_handler

    def _create_on_decision_handler(self, constraints: list, unhandled_obligations: list) -> \
            list[OnDecisionConstraintHandlerProvider]:
        """
        Creates a list with OnDecisionConstraintHandlerProvider, which are responsible for the given constraints

        :param constraints: Constraints to handle
        :param unhandled_obligations: list of all obligations, for which no responsible ConstraintHandlerProvider has been found
        :return: A list of OnDecisionConstraintHandlerProvider sorted by their priority, which are responsible for the given constraints
        """

        handler_list = []
        for constraint in constraints:
            for handler in self.on_decision_handler:
                self._add_responsible_handler(handler, constraint, handler_list, unhandled_obligations)
        handler_list.sort(key=lambda provider: provider.priority())
        return handler_list

    def _create_on_error_handler(self, constraints: list, unhandled_obligations: list) -> \
            list[ErrorConstraintHandlerProvider]:
        """
        Creates a list with ErrorConstraintHandlerProvider, which are responsible for the given constraints

        :param constraints: Constraints to handle
        :param unhandled_obligations: list of all obligations, for which no responsible ConstraintHandlerProvider has been found
        :return: A list of ErrorConstraintHandlerProvider sorted by their priority, which are responsible for the given constraints
        """

        handler_list = []

        for constraint in constraints:
            for handler in self.error_handler:
                self._add_responsible_handler(handler, constraint, handler_list, unhandled_obligations)
        handler_list.sort(key=lambda provider: provider.priority())
        return handler_list

    def _create_result_handler(self, constraints: list, unhandled_obligations: list) -> \
            list[ResultConstraintHandlerProvider]:
        """
        Creates a list with ResultConstraintHandlerProvider, which are responsible for the given constraints

        :param constraints: Constraints to handle
        :param unhandled_obligations: list of all obligations, for which no responsible ConstraintHandlerProvider has been found
        :return: A list of ResultConstraintHandlerProvider sorted by their priority, which are responsible for the given constraints
        """

        handler_list = []
        for constraint in constraints:
            for handler in self.result_handler:
                self._add_responsible_handler(handler, constraint, handler_list, unhandled_obligations)
        handler_list.sort(key=lambda provider: provider.priority())
        return handler_list

    def _create_function_argument_mapper(self, constraints: list, unhandled_obligations: list) -> \
            list[FunctionArgumentsConstraintHandlerProvider]:
        """
        Creates a list with FunctionArgumentsConstraintHandlerProvider, which are responsible for the given constraints

        :param constraints: Constraints to handle
        :param unhandled_obligations: list of all obligations, for which no responsible ConstraintHandlerProvider has been found
        :return: A list of FunctionArgumentsConstraintHandlerProvider sorted by their priority, which are responsible for the given constraints
        """

        handler_list = []
        for constraint in constraints:
            for handler in self.function_arguments_mapper:
                self._add_responsible_handler(handler, constraint, handler_list, unhandled_obligations)
        handler_list.sort(key=lambda provider: provider.priority())
        return handler_list

    @staticmethod
    def _add_responsible_handler(handler: ConstraintHandlerProvider, constraint, handler_list: list,
                                 unhandled_obligations: list) -> None:
        """
        Checks if the given ConstraintHandlerProvider is responsible to handle the given constraint. When the
        ConstraintHandlerProvider is responsible, it is added to a list of ConstraintHandlerProvider, which are used to
        create a ConstraintHandlerBundle. The constraint is removed from the list of obligations, which need to be
        handled

        :param handler: ConstraintHandlerProvider to check if it is responsible for the given constraint

        :param constraint: A constraint for the ConstraintHandlerProvider to check if it is responsible for it

        :param handler_list: A List containing all ConstraintHandlerProvider, which are responsible for given constraints

        :param unhandled_obligations: List of obligations, for which no responsible ConstraintHandlerProvider has been found so far
        """
        if handler.is_responsible(constraint):
            handler_list.append(handler)
            try:
                unhandled_obligations.remove(constraint)
            except ValueError:
                pass


constraint_handler_service = ConstraintHandlerService()
framework_permission_denied: ErrorConstraintHandlerProvider | None = None



from typing import Any, Callable

from sapl_base.constraint_handling.constraint_handler_provider import FunctionArgumentsConstraintHandlerProvider, \
    ErrorConstraintHandlerProvider, OnDecisionConstraintHandlerProvider, ResultConstraintHandlerProvider, \
    ConstraintHandlerProvider
from sapl_base.decision import Decision


class ConstraintHandlerBundle:
    """
    Contains lists with handle() methods of all ConstraintHandlerProvider, which are responsible for the obligations
    and advices of the provided Decision. The ConstraintHandlerBundle is created by the ConstraintHandlerService and
    is only created, when for every obligation is at least one ConstraintHandlerProvider responsible, otherwise an
    Exception is thrown.
    """


    def __init__(self,
                 on_decision_handler: list[OnDecisionConstraintHandlerProvider],
                 error_handler: list[ErrorConstraintHandlerProvider],
                 result_handler: list[ResultConstraintHandlerProvider],
                 function_arguments_mapper: list[FunctionArgumentsConstraintHandlerProvider] = None):
        """
        :param on_decision_handler: sorted List of OnDecisionConstraintHandlerProvider, of which their handle()
        methods are saved as a list[Callable[[Decision], None]]
        :param error_handler: sorted List of ErrorConstraintHandlerProvider, of which their handle()
        methods are saved as a list[Callable[[Exception], Exception]]
        :param result_handler: sorted List of ResultConstraintHandlerProvider, of which their handle()
        methods are saved as a list[Callable[[Any], Any]]
        :param function_arguments_mapper: sorted List of FunctionArgumentsConstraintHandlerProvider, of which their handle()
        methods are saved as a list[Callable[[dict], dict]]
        """

        if function_arguments_mapper is not None:
            self._function_arguments_mapper = self._add_handler_to_bundle(function_arguments_mapper)
        else:
            self._function_arguments_mapper = []

        self._on_decision_handler = self._add_handler_to_bundle(on_decision_handler)

        self._result_handler = self._add_handler_to_bundle(result_handler)

        self._error_handler = self._add_handler_to_bundle(error_handler)

    @staticmethod
    def _add_handler_to_bundle(handler_provider_list: list[ConstraintHandlerProvider]) -> list[Callable]:
        """
        Creates a list[Callable] of the handle() methods from the provided list[ConstraintHandlerProvider]

        :param handler_provider_list: List of ConstraintHandlerProvider, of which the handle() methods are added to a
        list[callable]

        :return: list of the handle() methods from the provided list[ConstraintHandlerProvider]
        """
        callable_list = []
        for handler in handler_provider_list:
            callable_list.append(handler.handle)
        return callable_list

    def execute_on_decision_handler(self, decision: Decision) -> None:
        """
        calls the handle() method of every responsible OnDecisionConstraintHandlerProvider with the given decision as
        argument.

        :param decision: a Decision received from the PolicyDecisionPoint
        """
        try:
            for handler in self._on_decision_handler:
                handler(decision)
        except Exception as e:
            raise self.execute_on_error_handler(e)

    def execute_on_error_handler(self, exception: Exception) -> None:
        """
        calls the handle() method of every responsible ErrorConstraintHandlerProvider with the given exception as
        argument.

        :param exception: an exception which was thrown, which will be handled by the ErrorConstraintHandlerProvider
        :raise Exception:
        """
        current_exception = exception
        for handler in self._error_handler:
            current_exception = handler(current_exception)
        raise current_exception

    def execute_result_handler(self, result: Any) -> Any:
        """
        calls the handle() method of every responsible ResultConstraintHandlerProvider with the given result as
        argument.

        :param result: the result from executing the enforced function
        :return: An object, which will be treated as the result of the enforced function
        """
        current_result = result
        try:
            for handler in self._result_handler:
                current_result = handler(current_result)
            return current_result
        except Exception as e:
            raise self.execute_on_error_handler(e)

    def execute_function_arguments_mapper(self, arguments: dict) -> dict:
        """
        calls the handle() method of every responsible FunctionArgumentsConstraintHandlerProvider with the given dict as
        argument.

        :param arguments: a dict containing the arguments, with which the enforced function will be called
        :return: a dict, with which the enforced function will be called
        """
        args = arguments
        try:
            for handler in self._function_arguments_mapper:
                args = handler(args)
            return args
        except Exception as e:
            raise self.execute_on_error_handler(e)



import asyncio
from functools import wraps

from sapl_base.policy_enforcement_points.streaming_policy_enforcement_point import StreamingPolicyEnforcementPoint
from sapl_base.policy_enforcement_points.async_policy_enforcement_point import AsyncPolicyEnforcementPoint
from sapl_base.policy_enforcement_points.sync_policy_enforcement_point import SyncPolicyEnforcementPoint
from sapl_base.sapl_util import double_wrap


@double_wrap
def pre_enforce(fn, subject: str = None, action: str = None, resource = None,
                environment: str = None, scope: str = "automatic"):
    """
    Wraps the decorated resource action point function with a SAPL Policy Enforcement Point (PEP).

    If the function returns a stream, pre_enforce will cancel the stream when a DENY is received.

    SAPL Decorators must be used as first Decorator to gather needed information of the annotated Function.

    Identify if the function is a coroutine and calls the appropriate function to handle the enforcement in a coroutine
    throws an exception, when the decorated function returns a stream and is no coroutine.

    :param fn:  decorated function
    :param subject: subject of an authorization_subscription or a function to create the subject
    :param action: action of an authorization_subscription or a function to create the action
    :param resource: resource of an authorization_subscription or a function to create the resource
    :param environment: environment of an authorization_subscription or a function to create the environment
    :param scope: Argument which creates a AuthorizationSubscription according to the given scope instead of evaluating the scope based on other parameter
    :return: The pre_enforced value of the decorated function
    """
    if asyncio.iscoroutinefunction(fn):
        @wraps(fn)
        async def async_wrap(*args, **kwargs):
            enforcement_point = AsyncPolicyEnforcementPoint(fn, *args, **kwargs)
            return await enforcement_point.pre_enforce(subject, action, resource, environment, scope)

        return async_wrap
    else:
        @wraps(fn)
        def wrap(*args, **kwargs):
            enforcement_point = SyncPolicyEnforcementPoint(fn, *args, **kwargs)
            return enforcement_point.pre_enforce(subject, action, resource, environment, scope)

        return wrap


@double_wrap
def post_enforce(fn, subject: str = None, action: str = None, resource: str = None,
                 environment: str = None, scope: str = "automatic"):
    """
    Post_enforces a decorated function with SAPL.

    SAPL Decorators must be used as first Decorator to gather needed information of the annotated Function.

    If the function returns a stream an exception is thrown.
    Identify if the function is a coroutine and calls the appropriate function to handle the enforcement in a coroutine

    :param fn:  decorated function
    :param subject: subject of an authorization_subscription or a function to create the subject
    :param action: action of an authorization_subscription or a function to create the action
    :param resource: resource of an authorization_subscription or a function to create the resource
    :param environment: environment of an authorization_subscription or a function to create the environment
    :param scope: Argument which creates a AuthorizationSubscription according to the given scope instead of evaluating the scope based on other parameter
    :return: The post_enforced value of the decorated function
    """
    if asyncio.iscoroutinefunction(fn):
        @wraps(fn)
        async def async_wrap(*args, **kwargs):
            enforcement_point = AsyncPolicyEnforcementPoint(fn, *args, **kwargs)
            return await enforcement_point.post_enforce(subject, action, resource, environment, scope)

        return async_wrap
    else:
        @wraps(fn)
        def wrap(*args, **kwargs):

            enforcement_point = SyncPolicyEnforcementPoint(fn, *args, **kwargs)
            return enforcement_point.post_enforce(subject, action, resource, environment, scope)

        return wrap


"""
SAPL Decorators must be used as first Decorator to gather needed information of the annotated Function.
"""


@double_wrap
def pre_and_post_enforce(fn, subject: str = None, action: str = None, resource: str = None,
                         environment: str = None, scope: str = "automatic"):
    """
    Pre- and post_enforces a decorated function with SAPL.

    SAPL Decorators must be used as first Decorator to gather needed information of the annotated Function.

    If the function returns a stream an exception is thrown.

    Identify if the function is a coroutine and calls the appropriate function to handle the enforcement in a coroutine

    :param fn:  decorated function
    :param subject: subject of an authorization_subscription or a function to create the subject
    :param action: action of an authorization_subscription or a function to create the action
    :param resource: resource of an authorization_subscription or a function to create the resource
    :param environment: environment of an authorization_subscription or a function to create the environment
    :param scope: Argument which creates a AuthorizationSubscription according to the given scope instead of evaluating the scope based on other parameter
    :return: The pre- and post_enforced value of the decorated function
    """
    if asyncio.iscoroutinefunction(fn):
        @wraps(fn)
        async def async_wrap(*args, **kwargs):
            enforcement_point = AsyncPolicyEnforcementPoint(fn, *args, **kwargs)
            return await enforcement_point.pre_and_post_enforce(subject, action, resource, environment, scope)

        return async_wrap
    else:
        @wraps(fn)
        def wrap(*args, **kwargs):
            enforcement_point = SyncPolicyEnforcementPoint(fn, *args, **kwargs)
            return enforcement_point.pre_and_post_enforce(subject, action, resource, environment, scope)

        return wrap


@double_wrap
def enforce_till_denied(fn, subject: str = None, action: str = None, resource: str = None,
                        environment: str = None, scope: str = "automatic"):
    """
    Enforces a stream and cancels the Stream when the Decision is not PERMIT

    SAPL Decorators must be used as first Decorator to gather needed information of the annotated Function.

    If the function doesn't return a stream, an exception will be thrown

    Identify if the function is a coroutine and will throw an exception, if it is not a coroutine.

    :param fn:  decorated function
    :param subject: subject of an authorization_subscription or a function to create the subject
    :param action: action of an authorization_subscription or a function to create the action
    :param resource: resource of an authorization_subscription or a function to create the resource
    :param environment: environment of an authorization_subscription or a function to create the environment
    :param scope: Argument which creates a AuthorizationSubscription according to the given scope instead of evaluating the scope based on other parameter
    :return: The return_value of the decorated function with a generator which will enforce each value, which is sent with the stream
    """
    if asyncio.iscoroutinefunction(fn):
        @wraps(fn)
        async def async_wrap(*args, **kwargs):
            enforcement_point = StreamingPolicyEnforcementPoint(fn, *args, **kwargs)
            return await enforcement_point.enforce_till_denied(subject, action, resource, environment, scope)

        return async_wrap
    else:
        @wraps(fn)
        def wrap(*args, **kwargs):
            raise Exception

        return wrap


@double_wrap
def enforce_drop_while_denied(fn, subject: str = None, action: str = None, resource: str = None,
                              environment: str = None, scope: str = "automatic"):
    """
    Enforces a stream and drops values, when the current decision is not PERMIT

    SAPL Decorators must be used as first Decorator to gather needed information of the annotated Function.

    If the function doesn't return a stream, an exception will be thrown

    Identify if the function is a coroutine and will throw an exception, if it is not a coroutine.

    :param fn:  decorated function
    :param subject: subject of an authorization_subscription or a function to create the subject
    :param action: action of an authorization_subscription or a function to create the action
    :param resource: resource of an authorization_subscription or a function to create the resource
    :param environment: environment of an authorization_subscription or a function to create the environment
    :param scope: Argument which creates a AuthorizationSubscription according to the given scope instead of evaluating the scope based on other parameter
    :return: The return_value of the decorated function with a generator which will enforce each value, which is sent with the stream
    """
    if asyncio.iscoroutinefunction(fn):
        @wraps(fn)
        async def async_wrap(*args, **kwargs):
            enforcement_point = StreamingPolicyEnforcementPoint(fn, *args, **kwargs)
            return await enforcement_point.drop_while_denied(subject, action, resource, environment, scope)

        return async_wrap
    else:
        @wraps(fn)
        def wrap(*args, **kwargs):
            raise Exception

        return wrap


@double_wrap
def enforce_recoverable_if_denied(fn, subject: str = None, action: str = None, resource: str = None,
                                  environment: str = None, scope: str = "automatic"):
    """
    Enforces a stream and drops values, when the current decision is not PERMIT and notify the client, that
    values are dropped, because they are not PERMITTED to receive them.

    SAPL Decorators must be used as first Decorator to gather needed information of the annotated Function.

    If the function doesn't return a stream, an exception will be thrown

    Identify if the function is a coroutine and will throw an exception, if it is not a coroutine.


    :param fn:  decorated function
    :param subject: subject of an authorization_subscription or a function to create the subject
    :param action: action of an authorization_subscription or a function to create the action
    :param resource: resource of an authorization_subscription or a function to create the resource
    :param environment: environment of an authorization_subscription or a function to create the environment
    :param scope: Argument which creates a AuthorizationSubscription according to the given scope instead of evaluating the scope based on other parameter
    :return: The return_value of the decorated function with a generator which will enforce each value, which is sent with the stream
    """
    if asyncio.iscoroutinefunction(fn):
        @wraps(fn)
        async def async_wrap(*args, **kwargs):
            enforcement_point = StreamingPolicyEnforcementPoint(fn, *args, **kwargs)
            return await enforcement_point.recoverable_if_denied(subject, action, resource, environment, scope)

        return async_wrap
    else:
        @wraps(fn)
        def wrap(*args, **kwargs):
            raise Exception

        return wrap
